#import "gdb.h"
#import "SegmentManagerThreads.h"
#import "threads.h"

int current_core_thread = 0;
static struct target_ops core_ops;

extern void child_attach (char *name, int from_tty);
extern void child_create_inferior (char *exec, char *args, char **env);
extern RegionManager *regionManager;
extern void core_get_registers(int regnum);
extern CORE_ADDR
getThreadPCAndNameFromState(struct REGS_STRUCT *regsState, char **pName);
extern char *
getThreadNameFromState(struct USER_REG_STRUCT *userRegState, int n);
extern void NeXT_add_file(char *arg_string, int from_tty);

static void
core_cleanup()
{
    if (regionManager) {
	[regionManager free];
	regionManager = nil;
    }
}

static void
core_open (char *filename, int from_tty)
{
    char *temp;
    struct cleanup *old_chain;
    SegmentManager *newCM;
  
    if (!filename) {
	error(regionManager
	      ? "No core file specified."
	        "  (Use `detach' to stop debugging a core file.)"
	      : "No core file specified.");
    }
    
    filename = tilde_expand (filename);
    if (filename[0] != '/') {
	temp = concat (current_directory, "/", filename);
	free (filename);
	filename = temp;
    }
    
    old_chain = make_cleanup (free, filename);
    newCM = [SegmentManager newCore: filename];
    if (!newCM)
	error("\"%s\" is not a core file.", filename);
	
    if (![newCM validate])
	error("\"%s\" is an invalid core file.", filename);
    
	
    discard_cleanups(old_chain);
    core_cleanup();
    regionManager = (RegionManager *)newCM;
    old_chain = make_cleanup(core_cleanup, NULL);
    
    push_target(&core_ops);
    core_get_registers(-1);
    set_current_frame(create_new_frame(read_register(FP_REGNUM), read_pc()));
    select_frame(get_current_frame(), 0);
    print_sel_frame(0);
    tell_client_program_stopped();
    
    discard_cleanups(old_chain);
}

static void
core_detach(char *args, int from_tty)
{
  dont_repeat ();
  if (args)
      error("Too many arguments");
  core_cleanup();
  if(from_tty)
    printf("No core file now.\n");
}

void
core_file_command(char *filename, int from_tty)
{
    if (!filename)
      core_detach(filename, from_tty);
    else
      core_open(filename, from_tty);
}

int
have_core_file_p ()
{
  return regionManager != NULL;
}

static void
core_files_info()
{
}

static int
core_xfer_memory (memaddr, myaddr, len, write)
     CORE_ADDR memaddr;
     char *myaddr;
     int len;
     int write;
{
    if (write)
	error("Can't write to core files.");
    return [regionManager getDataAt: (void *)memaddr for: len into: myaddr];
}

void
validate_files()
{
}

void
core_thread_list()
{
    int numThreads, i;
    ThreadInfo *tInfos, *tInfo;
    SegmentManager *coreManager = (SegmentManager *)regionManager;
    char *name, *fName;
    CORE_ADDR pc;
    
    numThreads = [coreManager numThreads];
    tInfos = [coreManager threadInfos];
    if (numThreads == 1)
	printf("There is 1 thread.\n");
    else
	printf("There are %d threads.\n", numThreads);
    printf("Thread PC         Name            Function\n");         
    for (i = 0; i < numThreads; i++) {
	name = getThreadNameFromState(tInfos[i].userReg, i);
	pc = getThreadPCAndNameFromState(tInfos[i].intRegs, &fName);
	printf ("%-6d 0x%-8x %-15s %-40s\n", i, pc, name, fName);
    }
    free(tInfos);
}

void
core_thread_select(char *args, int from_tty)
{
    int numThreads, numThread;
    SegmentManager *coreManager = (SegmentManager *)regionManager;
    
    if (!args)
	error_no_arg ("thread index to select");
    numThreads = [coreManager numThreads];
    numThread = atoi(args);
    if ((numThreads < numThreads) || (numThreads <= numThread))
	error("Invalid thread.\n");
    else
	current_core_thread = numThread;
}

static struct target_ops core_ops = {
    "core", "Local core dump file",
    "Use a core file as a target.  Specify the filename of the core file.",
    core_open, core_detach,
    child_attach, core_detach, 0, 0, /* resume, wait */
    core_get_registers,
    0, 0, 0, 0,
    core_xfer_memory, core_files_info,
    0, 0,
    0, 0, 0, 0, 0, /* terminal stuff */
    0, 0, NeXT_add_file, 0, 0, /* kill, add_file, call fn, lookup sym */
    child_create_inferior, 0, /* mourn_inferior */
    core_stratum, 0, /* next */
    0, 1, 1, 1, 0,	/* all mem, mem, stack, regs, exec */
    OPS_MAGIC,		/* Always the last thing */
};

void (*exec_file_display_hook) () = NULL;

void
_initialize_core()
{

  add_com ("core-file", class_files, core_file_command,
	   "Use FILE as core dump for examining memory and registers.\n\
No arg means have no core file.  This command has been superseded by the\n\
`target core' and `detach' commands.");
  add_target (&core_ops);
}
