/* This file is part of GDB.

GDB is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

GDB is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GDB; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

#import "gdb.h"
#include <errno.h>

#import <mach/cthreads.h>
#import <machine/reg.h>
#import <mach_oh.h>
#import <mach-o/nlist.h>

#import <mach-o/loader.h>

#import <objc/HashTable.h>
#import <objc/hashtable.h>

#import "loadSymbols.h"

task_t gdb_task;
cthread_t gdb_thread;
static NXZone *stringZone;
static NXHashTable *stringHash;

struct partial_symtab *mach_oh_psymtab_to_symtab();
void mach_oh_symfile_init();
char *find_opath_alias();

static struct sym_fns mach_oh_sym_fns = {
    "Mach_Oh",
    7,
    dbx_new_init,
    mach_oh_symfile_init,
    dbx_symfile_read,
    dbx_symfile_discard,
    mach_oh_psymtab_to_symtab
};

void
NeXT_add_file(char *arg_string, int from_tty)
{
    char *name;
    CORE_ADDR text_addr = 0;
    
    dont_repeat ();
    
    if (arg_string == 0)
	error ("add-symbol-file takes a file name and an optional address");
    
    arg_string = tilde_expand (arg_string);
    make_cleanup (free, arg_string);
    
    for(; *arg_string == ' '; arg_string++ );
    name = arg_string;
    
    for(; *arg_string && *arg_string != ' ' ; arg_string++ );
    *arg_string++ = '\0';
    
    if (*arg_string) {
	LoadInfo loadInfo;
	char *theNames[] = {name, NULL};
	
	loadInfo.headerAddr
	= (struct mach_header *)parse_and_eval_address(arg_string);
	loadSymbolsForInfo(&loadInfo, theNames);
    } else
	symbol_file_add(name, 0, 0, 0, exec_file_add(name, text_addr));
}

struct dbx_symfile_info {
  asection *text_sect;		/* Text section accessor */
  int symcount;			/* How many symbols are there in the file */
  char *stringtab;		/* The actual string table */
  int stringtab_size;		/* Its size */
  off_t symtab_offset;		/* Offset in file to symbol table */
  int desc;			/* File descriptor of symbol file */
};

void
mach_oh_symfile_init (sf)
  struct sym_fns *sf;
{
    bfd *sym_bfd = sf->sym_bfd;
    struct dbx_symfile_info *info;
    mach_oh_ex_data *ex_data = sym_bfd->tdata;
    
    /* Allocate struct to keep track of the symfile */
    sf->sym_private = xmalloc (sizeof (*info));
    info = (struct dbx_symfile_info *)sf->sym_private;
    
    //  info->desc = fileno ((FILE *)sym_bfd->iostream);
    info->desc = -1;
    info->text_sect = ex_data->textsec;
    if (!info->text_sect)
	abort();
    info->symcount = ex_data->n_syms;
    info->stringtab_size = ex_data->str_size;
    info->stringtab = ex_data->mapped_to + ex_data->str_filepos;
    
    info->symtab_offset = ex_data->sym_filepos;
}

void *
mach_oh_symtab(bfd *abfd, int *nbytes)
{
  mach_oh_ex_data *ex_data = abfd->tdata;
  *nbytes = ex_data->n_syms * sizeof(struct nlist);
  return ex_data->mapped_to + ex_data->sym_filepos;
}

struct partial_symtab *
mach_oh_psymtab_to_symtab (pst)
     struct partial_symtab *pst;
{
  mach_oh_ex_data *ex_data;
  extern bfd *symfile_bfd;
  extern int info_verbose;
  
  if (!pst)
    return;

  if (pst->symtab
  )
    {
      fprintf (stderr, "Psymtab for %s already read in.  Shouldn't happen.\n",
	       pst->filename);
      return;
    }

  if (pst->ldsymlen || pst->number_of_dependencies)
    {
      if (info_verbose)
        printf_filtered("Reading in symbols for %s...", pst->filename);
      symfile_bfd = pst->sf->sym_bfd;		/* Kludge for SWAP_SYMBOL */
      ex_data = symfile_bfd->tdata;
      /* FIXME, this uses internal BFD variables.  See above in
	 dbx_symbol_file_open where the macro is defined!  */
      psymtab_to_symtab_1 (pst,
			   -1,
//                           fileno((FILE *)symfile_bfd->iostream),
			   ex_data->mapped_to + ex_data->str_filepos,
			   ex_data->str_size,
			   ex_data->sym_filepos);

      /* Match with global symbols.  This only needs to be done once,
         after all of the symtabs and dependencies have been read in.   */
      scan_file_globals ();

      /* Finish up the debug error message.  */
      if (info_verbose)
	printf_filtered ("done.\n");
    }
}


#define N_TEXT 4
#define N_DATA 6
#define N_BSS 8

char *getShlibName(struct fvmlib_command *libCmd, char **originalName)
{
    char *name, *path, *subpath, *alias;
    int n;
    name = (void *)libCmd + libCmd->fvmlib.name.offset;
    if (path = (char *)find_opath_alias(name, &n)) {
	subpath = name + n;
	alias = xmalloc(strlen(path) + strlen(subpath) + 1);
	strcpy(alias, path);
	strcat(alias, subpath);
	if (originalName)
	    *originalName = name;
	return alias;
    } else {
	if (originalName)
	    *originalName = NULL;
	return name;
    }
}

BOOL static isExecutable(struct mach_header *header)
{
    return (header->filetype == MH_EXECUTE)
           || (header->filetype == MH_OBJECT)
           || (header->filetype == MH_PRELOAD);
}

void exec_file_add_shlibs(bfd *abfd)
{
    mach_oh_ex_data *ex_data = (mach_oh_ex_data *)abfd->tdata;
    struct mach_header *header = ex_data->mapped_to;
    int nCmds;
    struct load_command *loadCmd;
    char *name, *originalName;
    if (isExecutable(header)) {
        for (nCmds = header->ncmds,
	     loadCmd = (struct load_command *)(header + 1);
	     nCmds;
	     nCmds--, ((void *)loadCmd) += loadCmd->cmdsize) {
	    if (loadCmd->cmd == LC_LOADFVMLIB) {
		name = getShlibName((struct fvmlib_command *)loadCmd,
		                    &originalName);
                if (originalName)
		    printf("Aliasing %s to %s...\n", originalName, name);
		exec_file_add(name, 0);
	    }
	}
    }
}

void symbol_file_add_shlibs(bfd *abfd)
{
    mach_oh_ex_data *ex_data = (mach_oh_ex_data *)abfd->tdata;
    struct mach_header *header = ex_data->mapped_to;
    int nCmds;
    struct load_command *loadCmd;
    char *name;
    if (isExecutable(header)) {
        for (nCmds = header->ncmds,
	     loadCmd = (struct load_command *)(header + 1);
	     nCmds;
	     nCmds--, ((void *)loadCmd) += loadCmd->cmdsize) {
	    if (loadCmd->cmd == LC_LOADFVMLIB) {
		name = getShlibName((struct fvmlib_command *)loadCmd, NULL);
	        symbol_file_add(name, 1, 0, 0, NULL);
	    }
	}
    }
}

int
NeXT_validate_misc_function (CORE_ADDR address)
{
    bfd *e_bfd;
    mach_oh_ex_data *ex_data;
    for (e_bfd = exec_bfd; e_bfd; e_bfd = e_bfd->next) {
        ex_data = e_bfd->tdata;
        if (ex_data->flags.is_shlib
            && ((unsigned)ex_data->jmp_table_start <= address)
	    && (address < (unsigned)ex_data->jmp_table_end))
          return NO;
    }
    return YES;
}

#define MISC_BUNCH_SIZE 127

struct misc_bunch
{
  struct misc_bunch *next;
  struct misc_function contents[MISC_BUNCH_SIZE];
};

/* Bunch currently being filled up.
   The next field points to chain of filled bunches.  */

extern struct misc_bunch *misc_bunch;

/* Number of slots filled in current bunch.  */

extern int misc_bunch_index;

/* Total number of misc functions recorded so far.  */

extern int misc_count;

static int last_misc_count = 0;

void
start_psymtab_hook(struct partial_symtab *psymtab)
{
    last_misc_count = misc_count;
}

void
end_psymtab_hook(struct partial_symtab *psymtab)
{
    int count, index;
    struct misc_bunch *bunch;
    for (count = misc_count, bunch = misc_bunch, index = misc_bunch_index;
         count > last_misc_count;
	 count--) {
	if (!index) {
	    index = MISC_BUNCH_SIZE;
	    bunch = bunch->next;
	}
	index--;
	bunch->contents[index].usr_data = psymtab;
    }
}

typedef struct _MiscPSymtab {
    unsigned int val;
    struct partial_symtab *pst;
} MiscPSymtab;

static int miscIndex = 0;
static int numAlloced = 0;
#define INITIAL_ALLOC 500;
static MiscPSymtab *miscPSymtabs = NULL;

void
record_misc_function_psymtab(unsigned int val, struct partial_symtab *pst)
{
    if (!val || !NeXT_validate_misc_function(val))
        return;
    
    if (!numAlloced) {
	numAlloced = INITIAL_ALLOC;
	miscPSymtabs = xmalloc(numAlloced * sizeof(*miscPSymtabs));
    } else if (miscIndex == numAlloced) {
	numAlloced *= 2;
	miscPSymtabs = xrealloc(miscPSymtabs,
	                        numAlloced * sizeof(*miscPSymtabs));
    }
    miscPSymtabs[miscIndex].val = val;
    miscPSymtabs[miscIndex].pst = pst;
    miscIndex++;
}

int cmpMiscPSymtabs(const void *arg1, const void *arg2)
{
    const MiscPSymtab *mpst1 = arg1, *mpst2 = arg2;
    if (mpst1->val < mpst2->val)
	return -1;
    else if (mpst1->val == mpst2->val)
	return 0;
    else
	return 1;
}

void
combine_misc_functions_psymtabs()
{
    int count;
    MiscPSymtab *miscPSymtab;
    struct misc_function *miscFunction, *lastMiscFunction, *endOfMiscFunctions;
    
    if (miscIndex) {
	qsort(miscPSymtabs,
	      miscIndex,
	      sizeof(*miscPSymtabs),
	      &cmpMiscPSymtabs);
	miscFunction = misc_function_vector;
	endOfMiscFunctions = misc_function_vector + misc_function_count;
	for (count = miscIndex, miscPSymtab = miscPSymtabs;
	     count;
	     count--, miscPSymtab++) {
	    for (lastMiscFunction = miscFunction;
	         (miscFunction < endOfMiscFunctions)
		 && (miscFunction->address != miscPSymtab->val);
		 miscFunction++);
	    if (miscFunction < endOfMiscFunctions) {
		miscFunction->usr_data = miscPSymtab->pst;
		miscFunction++;
	    } else {
		miscFunction = lastMiscFunction;
		printf("Value with psymtab but no misc_function: %d in %s.\n",
		        miscPSymtab->val, miscPSymtab->pst->filename);
	    }
	}
	free (miscPSymtabs);
	miscPSymtabs = NULL;
	miscIndex = 0;
	numAlloced = 0;
    }
}

struct symtab *
symtab_for_misc_function(int ind)
{
    struct misc_function *misc_function;
    struct partial_symtab *psymtab;
    misc_function = misc_function_vector + ind;
    psymtab = misc_function->usr_data;
    if (psymtab)
	return PSYMTAB_TO_SYMTAB(psymtab);
    else
	return NULL;
}

static int
compare_blocks (const void *v1, const void *v2)
{
  const struct block * const *b1 = v1;
  const struct block * const *b2 = v2;
  
  if ((*b1)->startaddr < (*b2)->startaddr)
    return -1;
  else if ((*b1)->startaddr > (*b2)->startaddr)
    return 1;
  else if ((*b1)->endaddr < (*b2)->endaddr)
    return 1;
  else if ((*b1)->endaddr > (*b2)->endaddr)
    return -1;
}

void
make_block_vector_hook(struct blockvector *blockvector)
{
    /* We can no longer assume these will be sorted, so we sort them here.
       We don't sort the first two blocks, since their addresses are bogus. */
  
    if (BLOCKVECTOR_NBLOCKS (blockvector) > 2)
        qsort (&blockvector->block[2],
	       BLOCKVECTOR_NBLOCKS (blockvector) - 2,
	       sizeof (struct block *),
	       compare_blocks);
}

static int
compare_linetable_entries (lte1, lte2)
     struct linetable_entry *lte1, *lte2;
{
  register CORE_ADDR pc1 = lte1->pc;

  /* Return a signed result based on unsigned comparisons
     so that we sort into unsigned numeric order.  */
  if (pc1 < lte2->pc)
    return -1;
  return pc1 - lte2->pc;
}

void
linetable_hook(struct linetable *linetable)
{
    qsort (linetable->item, linetable->nitems,
	   sizeof (struct linetable_entry), compare_linetable_entries);
}

static HashTable *prologueHash = nil;

typedef union _ProEpInfo {
    unsigned int info;
    struct _Offsets {
	unsigned int prologue : 10;
	unsigned int epilogue : 22;
    } offsets;
} ProEpInfo;

void
mark_prologue(CORE_ADDR function, CORE_ADDR start, CORE_ADDR end)
{
    ProEpInfo proEpInfo;

    if (!prologueHash)
	prologueHash = [HashTable newKeyDesc: "i" valueDesc: "i"];

    proEpInfo.offsets.prologue = start - function;
    proEpInfo.offsets.epilogue =  end - function;
    
    [prologueHash insertKey: (void *)function value: (void *)proEpInfo.info];
}

CORE_ADDR
skip_prologue_by_symbol(function)
{
    CORE_ADDR prologue;
    ProEpInfo proEpInfo;
    
    proEpInfo.info = (unsigned)[prologueHash valueForKey: (void *)function];
    if (proEpInfo.info)
	return function + proEpInfo.offsets.prologue;
    else
	return 0;
}

CORE_ADDR
skip_epilogue(function)
{
    ProEpInfo proEpInfo;
    
    proEpInfo.info = (unsigned)[prologueHash valueForKey: (void *)function];
    if (proEpInfo.info)
	return (function + proEpInfo.offsets.epilogue);
    else
	return 0;
}

static void
end_command(char *args, int from_tty)
{
    ProEpInfo proEpInfo;
    struct symtab_and_line sal;
    CORE_ADDR function;
    struct block *block;
    
  if (args)
    error ("The \"end\" command does not take any arguments.");
  if (!target_has_execution)
    error ("The program is not running.");
  if (selected_frame == NULL)
    error ("No selected frame.");

    block = block_for_pc(stop_pc);
    function = block->startaddr;
    proEpInfo.info = (unsigned)[prologueHash valueForKey: (void *)function];
    if (proEpInfo.info) {
        if (stop_pc == function + proEpInfo.offsets.epilogue)
	    error("Already at end of function.");
        if (stop_pc > function + proEpInfo.offsets.epilogue)
	    error("Already past end of function.");
	sal = find_pc_line(function + proEpInfo.offsets.epilogue, 0);
	set_momentary_breakpoint (sal, selected_frame);
    } else
	error("Couldn't find end of function.");
    clear_proceed_status ();
    proceed (-1, -1, 0);
}

void
lock_execute_command(int lock)
{
    static mutex_t mutex = NULL;
    static cthread_t lockedThread = 0;
    if (!mutex)
	mutex = mutex_alloc();
    if (lock && (lockedThread != cthread_self())) {
	mutex_lock(mutex);
	lockedThread = cthread_self();
    } else {
	lockedThread = 0;
	mutex_unlock(mutex);
    }
}

void
execute_command_with_lock(char *p, int from_tty)
{
    struct cleanup *cleanups;
    lock_execute_command(YES);
    cleanups = make_cleanup(lock_execute_command, NO);
    execute_command(p, from_tty);
    do_cleanups(cleanups);
}
    
struct opaths {
  char *old, *new;
  struct opaths *next;
} *opaths;

opath_alias_command(paths)
char *paths;
{
  char *eqsign;
  char *index();
  struct opaths *opath;
  char *old, *new;

  if (!paths) {
    printf("Object File Path Aliases:\n");
    for(opath = opaths; opath; opath = opath->next) 
      printf("%s -> %s\n", opath->old, opath->new);
    return;
  }
  
  /* Should be of the form oldpath=newpath. */
  if (!(eqsign = index(paths, '='))) {
    fprintf(
      stderr,
      "Bad argument to -opath: %s, should be in format oldpath = newpath.\n",
      paths
    );
    return 0;
  }
  
  /* Skip leading whitespace */
  while(*paths == ' ' || *paths == '\t') paths++;
  old = paths;
  while (*paths != ' ' && *paths != '\t' && *paths != '=') paths++;
  *paths = 0;
  paths = eqsign+1;
  while(*paths == ' ' || *paths == '\t') paths++;
  new = paths;
  while (*paths != ' ' && *paths != '\t' && *paths != '\0') paths++;
  *paths = 0;

  opath = (struct opaths *)malloc(sizeof(struct opaths));
  opath->new = (char *)malloc(strlen(new) + 1);
  strcpy(opath->new, new);
  opath->old = (char *)malloc(strlen(old) + 1);
  strcpy(opath->old, old);
  opath->next = opaths;
  opaths = opath;
}

static
opath_unalias_command (path)
char *path;
{
  struct opaths *opath;

  if (!path)
    error_no_arg ("old pathname");
  /* Skip leading whitespace */
  while(*path == ' ' || *path == '\t') path++;

  if (opaths) {
    if (!strcmp(opaths->old, path)) {
      opaths = opaths->next;
      return;
    } else {
      for(opath = opaths; opath->next; opath = opath->next) {
	if (!strcmp(opath->next->old, path)) {
	  struct opaths *o;
	  free(opath->next->old);
	  free(opath->next->new);
	  o = opath->next->next;
	  free(opath->next);
	  opath->next = o;
	  return;
	}
      }
    }
  }
  fprintf(stderr, "Can't find %s in opath alias list.\n", path);
}

char *
find_opath_alias(string, ncharsubst)
char *string;
int *ncharsubst;
{
  struct opaths *opath;

  /* Do O-path aliasing. */
  for(opath = opaths; opath; opath = opath->next) {
    if (!strncmp(opath->old, string, strlen(opath->old))) {
      char *newname, *subpath;
      
      *ncharsubst = strlen(opath->old);
      return(opath->new);
    }
  }
  return 0;
}

char *
stringSave(char *string)
{
    char *savedString;
    int n;
    
    if (!string)
	return NULL;

    if (savedString = NXHashGet(stringHash, string))
	return savedString;
    else {
	n = strlen(string);
	savedString = NXZoneMalloc(stringZone, n + 1);
	bcopy(string, savedString, n);
	savedString[n] = '\0';
	NXHashInsert(stringHash, savedString);
	return savedString;
    }
}

char *
stringSave2(char *s1, char *s2)
{
    char s3[strlen(s1) + strlen(s2) + 1];
    
    strcpy(s3, s1);
    strcat(s3, s2);
    return stringSave(s3);
}

char *
stringSaveN(char *string, int n)
{
    char *savedString;
    char newString[n + 1];

    if (!string)
	return NULL;

    bcopy(string, newString, n);
    newString[n] = '\0';
    if (savedString = NXHashGet(stringHash, newString))
	return savedString;
    else {
	savedString = NXZoneMalloc(stringZone, n + 1);
	bcopy(string, savedString, n);
	savedString[n] = '\0';
	NXHashInsert(stringHash, savedString);
	return savedString;
    }
}

int
stringEq(char *s1, char *s2)
{
    if ((NXZoneFromPtr(s1) == stringZone) && (NXZoneFromPtr(s2) == stringZone))
	return s1 == s2;
    else
	return (strcmp(s1, s2) == 0);
}

static int pstrcmp(const void *arg1, const void *arg2)
{ return strcmp(*(char **)arg1, *(char **)arg2); }

static void
dumpStrings(char *args)
{
    FILE *stringsFile = fopen(args, "w+");
    int count = NXCountHashTable(stringHash), i = 0;
    NXHashState state = NXInitHashState(stringHash);
    char *strings[sizeof(char *) * count], *string;

    while (NXNextHashState(stringHash, &state, &(void *)string)) {
	strings[i++] = string;
    }
    qsort(strings, count, sizeof(*strings), pstrcmp);
    for (i = 0; i < count; i++) {
	fprintf(stringsFile, "%s\n", strings[i]);
    }
    fclose(stringsFile);
}

static void
dumpMisc(char *args)
{
    FILE *miscFile = fopen(args, "w+");
    int count, numDups = 0;
    struct misc_function *mf;

    for (count = misc_function_count, mf = misc_function_vector;
         count > 0;
	 count--, mf++) {
	if ((count - 1) && (mf->name == (mf + 1)->name) && (mf->address == (mf + 1)->address)) {
	    fprintf(miscFile, "%s\t%p\n", mf->name, mf->address);
	    mf++;
	    count--;
	    numDups++;
	}
    }
    fprintf(miscFile, "Number of duplicates: %d.\n", numDups);
    fclose(miscFile);
}

void
_initialize_string()
{
    static BOOL initialized = NO;
    if (!initialized) {
	stringHash = NXCreateHashTable(NXStrPrototype, 0, NULL);
	stringZone = NXCreateZone(vm_page_size, vm_page_size, NO);
	NXNameZone(stringZone, "StringZone");
	initialized = YES;
    }
}
	
void
memory_error (status, memaddr)
     int status;
     CORE_ADDR memaddr;
{

  if (status == EIO)
    {
      /* Actually, address between memaddr and memaddr + len
	 was out of bounds. */
      error ("Cannot access memory: address 0x%x out of bounds.", memaddr);
    }
  else
    {
      if (status >= sys_nerr || status < 0)
	error ("Error accessing memory address 0x%x: unknown error (%d).",
	       memaddr, status);
      else
	error ("Error accessing memory address 0x%x: %s.",
	       memaddr, sys_errlist[status]);
    }
}

/* Same as target_read_memory, but report an error if can't read.  */
void
read_memory (memaddr, myaddr, len)
     CORE_ADDR memaddr;
     char *myaddr;
     int len;
{
  int status;
  status = target_read_memory (memaddr, myaddr, len);
  if (status != 0)
    memory_error (status, memaddr);
}

/* Same as target_write_memory, but report an error if can't write.  */
void
write_memory (memaddr, myaddr, len)
     CORE_ADDR memaddr;
     char *myaddr;
     int len;
{
  int status;

  status = target_write_memory (memaddr, myaddr, len);
  if (status != 0)
    memory_error (status, memaddr);
}

/* Read an integer from debugged memory, given address and number of bytes.  */

long
read_memory_integer (memaddr, len)
     CORE_ADDR memaddr;
     int len;
{
  char cbuf;
  short sbuf;
  int ibuf;
  long lbuf;

  if (len == sizeof (char))
    {
      read_memory (memaddr, &cbuf, len);
      return cbuf;
    }
  if (len == sizeof (short))
    {
      read_memory (memaddr, (char *)&sbuf, len);
      SWAP_TARGET_AND_HOST (&sbuf, sizeof (short));
      return sbuf;
    }
  if (len == sizeof (int))
    {
      read_memory (memaddr, (char *)&ibuf, len);
      SWAP_TARGET_AND_HOST (&ibuf, sizeof (int));
      return ibuf;
    }
  if (len == sizeof (lbuf))
    {
      read_memory (memaddr, (char *)&lbuf, len);
      SWAP_TARGET_AND_HOST (&lbuf, sizeof (lbuf));
      return lbuf;
    }
  error ("Cannot handle integers of %d bytes.", len);
  return -1;	/* for lint */
}

void 
_initialize_next()
{
    add_symtab_fns(&mach_oh_sym_fns);
    add_com("end", class_breakpoint, end_command,
            "Execute until end of current function.");
    add_com ("opath-alias", class_obscure, opath_alias_command,
	    "Add path alias for lazy symbol lookup from .o file. OLDPATH=NEWPATH");
    add_com_alias ("opath", "opath-alias", class_obscure, 1);
    add_com ("opath-unalias", class_obscure, opath_unalias_command,
	    "Delete path alias for symbol lookup from .o file. OLDPATH");
#if 0
    add_com ("dump-strings", class_obscure, dumpStrings,
             "Dump all the strings seen into FILE.");	    
    add_com ("dump-misc", class_obscure, dumpMisc,
             "Dump all the strings seen into FILE.");	    
#endif 0	     
    gdb_task = task_self();
    gdb_thread = cthread_self();
}