#import "gdb.h"
#import "objc.h"
#import "mach_oh.h"

#import <mach-o/ldsyms.h>
#import <stdio.h>
#import <stdlib.h>

#import <mach-o/loader.h>

#import <objc/hashtable.h>
#import <objc/maptable.h>

extern RegionManager *regionManager;

typedef struct _ClassImp {
    char *className;
    IMP function;
    BOOL isFactory;
    struct _ClassImp *next;
} ClassImp;

static NXHashTable *selectorHash;
static NXMapTable *classMap;
static Class stringClass = NULL;

typedef enum _OnlyOne {ANYTHING, INSTANCEONLY, CLASSONLY} OnlyOne;

static inline BOOL streq(char *str1, char *str2)
{
    return ((*str1 == *str2) && (strcmp(str1, str2) == 0));
}

#include <sys/time.h>
#include <sys/resource.h>
#include <time.h>

/* Return time used so far, in microseconds.  */

int gettime ()
{
  struct rusage rusage;

  getrusage (0, &rusage);
  return (rusage.ru_utime.tv_sec * 1000000 + rusage.ru_utime.tv_usec
	  + rusage.ru_stime.tv_sec * 1000000 + rusage.ru_stime.tv_usec);
}

void print_time (str, total, tv)
     char *str;
     int total;
     struct timeval *tv;
{
  extern char **NXArgv;

  fprintf (stderr,
	   "time to `%s' for `%s':\n\t%7d.%06d (sys+usr)"
	   	"\t%7d.%01d (real).\n", str, NXArgv[0],
	   total / 1000000, total % 1000000, 
		tv->tv_sec, tv->tv_usec/100000);
}

void *warpAddr(bfd *aBfd, const void *addr)
{
    mach_oh_ex_data *machOhData = aBfd->tdata;
    struct gdb_ex_data *gdbData = aBfd->usrdata;
    struct section_table *section, *foundSection;
    for (foundSection = NULL, section = gdbData->sections;
         !foundSection && section < gdbData->sections_end;
	 section++) {
	if ((section->addr <= (unsigned)addr)
	                  && ((unsigned)addr < section->endaddr))
	    foundSection = section;
    }
    if (foundSection)
        return (machOhData->mapped_to
                + (((unsigned)addr - foundSection->addr)
	            + foundSection->sec_ptr->filepos));
    else
	return NULL;
}

static inline
ClassImp *findImp(bfd *aBfd, char *methodName, char *className, BOOL isFactory,
                  struct objc_method_list *methods, ClassImp *imps)
{
    Method foundMethod = NULL, method;
    int nMethods;
    ClassImp *imp;
    for (nMethods = methods->method_count, method = methods->method_list;
         !foundMethod && nMethods;
	 nMethods--, method++) {
	if (streq(warpAddr(aBfd, method->method_name), methodName))
	    foundMethod = method;
    }
    if (foundMethod) {
	imp = xmalloc(sizeof(*imp));
	imp->className = className;
	imp->function = foundMethod->method_imp;
	imp->isFactory = isFactory;
	imp->next = imps;
	return imp;
    } else
	return imps;
}

static ClassImp *
findClassImps(bfd *abfd, char *methodName, char *className, OnlyOne onlyOne,
              Class *pClass, int nClasses, ClassImp *imps)
{
    Class class, metaClass;
    char *clsName;
    struct objc_method_list *methods;
    
    for (; nClasses; nClasses--, pClass++) {
	class = warpAddr(abfd, *pClass);
	clsName = warpAddr(abfd, class->name);
	if (!className || (strcmp(clsName, className) == 0)) {
	    if ((onlyOne != CLASSONLY)
		&& (methods = warpAddr(abfd, class->methods)))
		imps = findImp(abfd, methodName, clsName, NO, methods, imps);
	    if ((onlyOne != INSTANCEONLY)
	        && (metaClass = warpAddr(abfd, class->isa))
		&& (methods = warpAddr(abfd, metaClass->methods)))
		imps = findImp(abfd, methodName, clsName, YES, methods, imps);
	    if (onlyOne && imps)
		return imps;
	}
    }
    return imps;
}

static ClassImp *
findCatImps(bfd *abfd, char *methodName, char *className, OnlyOne onlyOne,
            Category *pCat, int nCats, ClassImp *imps)
{
    Category cat;
    char *clsName;
    struct objc_method_list *methods;

    for (; nCats; nCats--, pCat++) {
	cat = warpAddr(abfd, *pCat);
	clsName = warpAddr(abfd, cat->class_name);
	if (!className || (strcmp(clsName, className) == 0)) {
	    if ((onlyOne != CLASSONLY)
		&& (methods = warpAddr(abfd, cat->instance_methods)))
		imps = findImp(abfd, methodName, clsName, NO, methods, imps);
	    if ((onlyOne != INSTANCEONLY)
		&& (methods = warpAddr(abfd, cat->class_methods)))
		imps = findImp(abfd, methodName, clsName, NO, methods, imps);
	    if (onlyOne && imps)
		return imps;
	}
    }
    return imps;
}

static ClassImp *
findImps(char *methodName, char *className, OnlyOne onlyOne)
{
    ClassImp *imps = NULL;
    bfd *eBfd;
    Module mod;
    int nMods;
    Symtab symtab;
    asection *asect;
    
    for (eBfd = exec_bfd; eBfd; eBfd = eBfd->next) {
        if (asect = bfd_get_section_by_name(eBfd, "__OBJC__module_info")) {
	    for (nMods = asect->size / sizeof(struct objc_module),
		 mod = warpAddr(eBfd, (void *)bfd_section_vma(abfd, asect));
		 nMods;
		 nMods--, mod++) {
		symtab = warpAddr(eBfd, mod->symtab);
		if (symtab) {
		    imps = findClassImps(eBfd, methodName, className, onlyOne,
		                         (Class *)symtab->defs, 
					 symtab->cls_def_cnt,
					 imps);
		    if (onlyOne && imps)
			return imps;
		    imps = findCatImps(eBfd, methodName, className, onlyOne,
		                       (Category *)symtab->defs
				                   + symtab->cls_def_cnt,
				       symtab->cat_def_cnt,
				       imps);
		    if (onlyOne && imps)
			return imps;
		}
	    }
	}
    }
    return imps;
}

Class
lookup_objc_class(char *className)
{
    if (classMap && className && *className)
	return NXMapGet(classMap, className);
}

BOOL
isClass(Class class, BOOL allowMeta)
{
    bfd *eBfd;
    unsigned firstClass;
    asection *sect;
    
    for (eBfd = exec_bfd; eBfd; eBfd = eBfd->next) {
        if (sect = bfd_get_section_by_name(eBfd, "__OBJC__class")) {
	    firstClass = bfd_section_vma(eBfd, sect);
	    if ((firstClass <= (unsigned)class) 
	        && ((unsigned)class < (firstClass + sect->size)))
		return YES;
	}
	if (allowMeta) {
	    if (sect = bfd_get_section_by_name(eBfd, "__OBJC__meta_class")) {
		firstClass = bfd_section_vma(eBfd, sect);
		if ((firstClass <= (unsigned)class) 
			&& ((unsigned)class < (firstClass + sect->size)))
		    return YES;
	    }
	}
    }
    if (!stringClass) {
	int index = lookup_misc_func("_NXConstantString_class");
	if (index != -1)
	    stringClass = (Class)misc_function_vector[index].address;
    }
    if (stringClass == class)
	return YES;
    return NO;
}

BOOL
isMetaClass(Class class)
{
    bfd *eBfd;
    unsigned firstClass;
    asection *sect;
    
    for (eBfd = exec_bfd; eBfd; eBfd = eBfd->next) {
        if (sect = bfd_get_section_by_name(eBfd, "__OBJC__meta_class")) {
	    firstClass = bfd_section_vma(eBfd, sect);
	    if ((firstClass <= (unsigned)class) 
	                   && ((unsigned)class < (firstClass + sect->size)))
		return YES;
	}
    }
    return NO;
}

int
lookup_child_selector(char *methodName)
{
    value function, val, methodString;
    int methodSel;
    
    if (selectorHash
        && methodName && *methodName
	&& NXHashMember(selectorHash, methodName)) {
	function = value_of_function("sel_getUid");
	if (function) {
	    methodString = value_string(methodName, strlen(methodName));
	    val = target_call_function(function, 1, &methodString);
	    if (value_zerop(val))
		error("No method named %s.", methodName);
	    else 
		return value_as_long(val);
	} else
	    error("Couldn't find sel_getUid in inferior.");
    } else
	error("No method named %s.", methodName);
    return 0;
}

char *
lookup_child_selector_name(SEL sel)
{
    value function, val, selector;
    char *methodName;
    
    function = value_of_function("sel_getName");
    if (function) {
	selector = value_from_long(builtin_type_long, sel);
	val = target_call_function(function, 1, &selector);
	if (value_zerop(val))
	    error("No method named %s.", methodName);
	else 
	    return WARPSTRING((char *)value_as_long(val));
    } else
	error("Couldn't find sel_getName in inferior.");
    return 0;
}

Class
classForObject(id object)
{
    id myObject = WARP(object);
    Class class;
    
    if (myObject && isClass(class = myObject->isa, YES))
	return class;
    else
	return NULL;
}

Class
classForObjectOnly(id object)
{
    id myObject = WARP(object);
    Class class;
    
    if (myObject && isClass(class = myObject->isa, NO))
	return class;
    else
	return NULL;
}

CORE_ADDR
find_implementation_from_class(Class class, SEL sel)
{
    Class myClass;
    Method method, foundMethod = NULL;
    struct objc_method_list *methods;
    int nMethods;
    for (myClass = WARP(class);
	 !foundMethod && myClass;
	 myClass = WARP(myClass->super_class)) {
	for (foundMethod = NULL, methods = WARP(myClass ->methods);
	     !foundMethod && methods;
	     methods = WARP(methods->method_next)) {
	    for (nMethods = methods->method_count,
		 method = methods->method_list;
		 !foundMethod && nMethods;
		 nMethods--, method++) {
		if (method->method_name == sel)
		    foundMethod = method;
	    }
	}
    }
    if (foundMethod)
	return (CORE_ADDR)foundMethod->method_imp;
    else
	return 0;
}

CORE_ADDR
find_implementation(id object, SEL sel)
{
    Class class;
    if (class = classForObject(object))
	return find_implementation_from_class(class, sel);
    else
	return 0;
}

int cmpImps(const void *i1, const void *i2)
{
    ClassImp * const *imp1 = i1, * const *imp2 = i2;
    return strcmp((*imp1)->className, (*imp2)->className);
}

int
lookup_method(char **argPtr, struct symtab_and_line *val)
{
    int ret = 0, impIndex, i, size, j, maxLen = 0, maxNum, len, numPerLine;
    int num, numImps, numLines, numFullLines, line;
    char buf[80], *p = *argPtr, *saveP, *methodName, *className;
    ClassImp *imps, *imp, **impArray;
    OnlyOne onlyOne = ANYTHING;
    struct misc_function *miscFunc = NULL;
    CORE_ADDR pc = 0;
    
    if (strchr("+-[", *p)) {
	switch (*p) {
	    case '-':
		onlyOne = INSTANCEONLY;
		if (*(p + 1) == '[') 
		    p += 2;
		else
		    return 0;
		break;
	    case '+':
		onlyOne = CLASSONLY;
		if (*(p + 1) == '[') 
		    p += 2;
		else
		    return 0;
		break;
	    case '[': default:
		p++;
		break;
	}

	/* Copy class name */
	for (; *p && *p == ' '; p++);
	for (saveP = p, size = 0; *p && (*p != ' '); p++, size++);
	className = alloca(size + 1);
	memcpy(className, saveP, size), className[size] = '\0';
	
	/* Copy method name */
	for (; *p && *p == ' '; p++);
	for (saveP = p, size = 0;
	     *p && (*p != ' ') && (*p != ']');
	     p++, size++);
	methodName = alloca(size + 1);
	memcpy(methodName, saveP, size), methodName[size] = '\0';
	if (*p == ']')
	    p++;
    } else {
    
	/* Copy method name */
	for (; *p && *p == ' '; p++);
	for (saveP = p, size = 0; *p && (*p != ' '); p++, size++);
	methodName = alloca(size + 1);
	memcpy(methodName, saveP, size), methodName[size] = '\0';

	className = NULL;
    }
    if (selectorHash 
        && methodName && *methodName && NXHashMember(selectorHash, methodName)
        && (imps = findImps(methodName, className, onlyOne))) {
	if (!className && !onlyOne)
	    miscFunc = lookup_misc_function(methodName);
	if (imps->next || miscFunc) {
	    for (imp = imps, numImps = 0; imp; imp = imp->next, numImps++);
	    impArray = alloca(numImps * sizeof(*impArray));
	    for (imp = imps, i = 0; imp; imp = imp->next, i++) {
		impArray[i] = imp;
		len = strlen(imp->className);
		if (len > maxLen)
		    maxLen = len;
	    }
	    qsort(impArray, numImps, sizeof(*impArray), cmpImps);
	    for (maxNum = 1, num = 10; num <= numImps; maxNum++, num *= 10);
	    numPerLine = 80 / (maxLen + maxNum + 4);
	    if (numPerLine < 1) numPerLine = 1;
	    if (numPerLine > i) numPerLine = numImps;
	    maxLen
	    += ((80 - (numPerLine * (maxLen + maxNum + 4))) / numPerLine);
	    numLines = ((numImps - 1) / numPerLine) + 1;
	    numFullLines = numImps - ((numImps / numPerLine) * numPerLine);
		    
	    if (numImps == 1)
		printf_filtered("The following class implements %s:\n",
		                methodName);
	    else
		printf_filtered("The following classes implement %s:\n",
		                methodName);
	    for (i = 0, line = 0; i < numImps; line++) {	  
		for (j = 0; i < numImps && j < numPerLine; j++, i++) {
		    if (!numFullLines || j <= numFullLines)
			impIndex = (j * numLines) + line;
		    else
			impIndex = (numFullLines * numLines)
			           + ((j - numFullLines) * (numLines - 1))
				   + line;
		    imp = impArray[impIndex];
		    printf_filtered("%*d) %c%-*s ",
		     		    maxNum,
				    impIndex + 1,
		                    imp->isFactory ? '+' : '-',
				    maxLen,
				    imp->className);
		}
		printf_filtered("\n");
	    }
	    if (miscFunc) {
		if (numImps > 1)
		    printf_filtered("\n");
		printf_filtered("%*d) %s() as a function\n\n",
		                maxNum, numImps + 1, methodName);
	    }
	    printf_filtered("Which one do you want? ");
	    fgets(buf, sizeof(buf), stdin);
	    if (isdigit(buf[0])) {
		impIndex = atoi(buf);
		if ((1 <= impIndex) && (impIndex <= numImps))
		    pc = (CORE_ADDR)impArray[impIndex - 1]->function;
		else if (miscFunc && (impIndex == (numImps + 1)))
		    pc = miscFunc->address;
	    } else {
		char *retChar = (char *)index(buf, '\n');
		
		if (retChar)
		    *retChar = '\0';
		for (i = 0; i < numImps && !pc; i++) {
		    if (strcmp(buf, impArray[i]->className) == 0)
			pc = (CORE_ADDR)impArray[i]->function;
		}
	    }
	    for (i = 0; i < numImps; i++)
		free(impArray[i]);
	} else
            pc = (CORE_ADDR)imps->function;
	if (pc) {
	    find_pc_symtab(pc);
	    SKIP_PROLOGUE(pc);
	    *val = find_pc_line(pc, 0);
	    if (!val->symtab)
		val->pc = pc;
	    ret = 1;
	}
	if (!pc)
	    error("No method chosen.");
    }
    if (ret) {
	*argPtr = p;
	return ret;
    } else if (className) {
	if (methodName && *methodName) {
	    printf_filtered("%s responds to no ", className);
	    if (onlyOne == CLASSONLY)
		printf_filtered("class ");
	    else if (onlyOne == INSTANCEONLY)
		printf_filtered("instance ");
	    error("method named %s.", methodName);
	} else
	    error ("No method name given.");
    } else
	return 0;
}

void static inline
addMethods(bfd *aBfd, struct objc_method_list *methods)
{
    Method method;
    int nMethods;
    
    if (methods) {
	for (nMethods = methods->method_count, method = methods->method_list;
	     nMethods;
	     nMethods--, method++) {
	    NXHashInsertIfAbsent(selectorHash,
				    warpAddr(aBfd, method->method_name));
	}
    }
}

void
objc_completion_list_hook (char *text)
{
    int nameLen;
    char firstChar, *selector;
    NXHashState state;
    
    if (!selectorHash)
	return;
    nameLen = strlen(text);
    if (nameLen) {
	nameLen--;
	firstChar = *text;
	text++;
	state = NXInitHashState(selectorHash);
	while (NXNextHashState(selectorHash, &state, &(void *)selector)) {
	    if ((firstChar == *selector)
	        && (strncmp(text, selector + 1, nameLen) == 0))
		completion_list_add_symbol(selector);
	}
    }
}

struct type *
typeForObject(CORE_ADDR object)
{
    Class class, myClass;
    char *myClassName;
    struct type *structType;
    
    if (regionManager
    	&& (class = classForObjectOnly((id)object))
	&& (myClass = WARP(class))
	&& (myClassName = WARPSTRING(myClass->name))
	&& (structType = lookup_struct_no_error(myClassName, 0)))
	return lookup_pointer_type(structType);
    else
	return NULL;
}

struct type *
typeForClassName(char *className)
{
    if (NXMapGet(classMap, className))
        return lookup_struct_no_error(className, 0);
    else
	return NULL;
}

int
isObjectType(struct type *type)
{
    struct type *target_type;
    char *name;
    
    return ((TYPE_CODE(type) == TYPE_CODE_PTR)
            && (target_type = TYPE_TARGET_TYPE(type))
	    && (TYPE_CODE(target_type) == TYPE_CODE_STRUCT)
	    && (name = TYPE_NAME(target_type))
	    && (strncmp(name, "struct ", 7) == 0)
	    && (lookup_objc_class(name + 7)
	        || (strcmp(name + 7, "objc_object") == 0)));
}

struct type *
objectType(CORE_ADDR object)
{
    Class class, myClass;
    char *myClassName;
    
    if (regionManager
        && (class = classForObject((id)object))
	&& (!isMetaClass(class))
	&& (myClass = WARP(class))
	&& (myClassName = WARPSTRING(myClass->name)))
	return lookup_struct_no_error(myClassName, 0);
    else
	return NULL;
}

static void
addModSect(bfd *abfd, asection *asect)
{
    Class class, *pClass;
    int nClasses, nMods, nCats;
    Module mod;
    Symtab symtab;
    Category cat, *pCat;
    
    for (nMods = asect->size / sizeof(struct objc_module),
         mod = warpAddr(abfd, (void *)bfd_section_vma(abfd, asect));
         nMods;
	 nMods--, mod++) {
	symtab = warpAddr(abfd, mod->symtab);
	if (symtab) {
	    for (nClasses = symtab->cls_def_cnt,
	         pClass = (Class *)symtab->defs;
		 nClasses;
		 nClasses--, pClass++) {
		class = warpAddr(abfd, *pClass);
		NXMapInsert(classMap, warpAddr(abfd, class->name), *pClass);
		addMethods(abfd, warpAddr(abfd, class->methods));
		class = warpAddr(abfd, class->isa);
		addMethods(abfd, warpAddr(abfd, class->methods));
	    }
	    for (nCats = symtab->cat_def_cnt, pCat = (Category *)pClass;
		 nCats;
		 nCats--, pCat++) {
		cat = warpAddr(abfd, *pCat);
		addMethods(abfd, warpAddr(abfd, cat->instance_methods));
		addMethods(abfd, warpAddr(abfd, cat->class_methods));
	    }
	}
    }
}

void
readObjC(bfd *eBfd)
{
    asection *modSect;

    if (!selectorHash)
	selectorHash = NXCreateHashTable(NXStrPrototype, 0, NULL);
    if (!classMap)
	classMap = NXCreateMapTable(NXStrValueMapPrototype, 0);

    if (modSect = bfd_get_section_by_name(eBfd, "__OBJC__module_info"))
	addModSect(eBfd, modSect);
}

struct type *
objc_type(value val)
{
    struct type *oldType, *newType;
    
    oldType = VALUE_TYPE(val);
    if (isObjectType(oldType))
	return typeForObject(*(CORE_ADDR *)VALUE_CONTENTS(val));
    else
	return NULL;
}

void
cast_to_objc_type(value val)
{
    struct type *newType;
    
    if (newType = objc_type(val))
        VALUE_TYPE(val) = newType;
}

value
value_of_self (complain)
     int complain;
{
  extern FRAME selected_frame;
  struct symbol *func, *sym;
  struct block *b;
  static char *savedSelf = NULL;
  
  if (selected_frame == 0)
    if (complain)
      error ("no frame selected");
    else return 0;

  func = get_frame_function (selected_frame);
  if (!func)
    return 0;
    
  b = SYMBOL_BLOCK_VALUE (func);
  if (BLOCK_NSYMS (b) <= 0)
    if (complain)
      error ("no args, no `self'");
    else return 0;

  sym = BLOCK_SYM (b, 0);

  if (!savedSelf)
    savedSelf = stringSave("self");

  if (savedSelf != SYMBOL_NAME (sym))
    if (complain)
      error ("current stack frame not in method");
    else return 0;
    
  sym = lookup_block_symbol(b, savedSelf, VAR_NAMESPACE);
      
  return read_var_value (sym, selected_frame);
}

void static
removeMethods(bfd *aBfd, struct objc_method_list *methods)
{
    Method method;
    int nMethods;
    char *sel, *theSel;
    
    if (methods) {
	for (nMethods = methods->method_count, method = methods->method_list;
	     nMethods;
	     nMethods--, method++) {
	    sel = warpAddr(aBfd, method->method_name);
	    if (sel == NXHashGet(selectorHash, sel)) {
		if (aBfd->next) {
		    theSel = savestring(sel, strlen(sel));
		    NXHashInsert(selectorHash, theSel);
		} else
		    NXHashRemove(selectorHash, sel);
	    }
	}
    }
}

static void
removeModSect(bfd *abfd, asection *asect)
{
    Class class, *pClass;
    int nClasses, nMods, nCats;
    Module mod;
    Symtab symtab;
    Category cat, *pCat;
    
    for (nMods = asect->size / sizeof(struct objc_module),
         mod = warpAddr(abfd, (void *)bfd_section_vma(abfd, asect));
         nMods;
	 nMods--, mod++) {
	symtab = warpAddr(abfd, mod->symtab);
	if (symtab) {
	    for (nClasses = symtab->cls_def_cnt,
	         pClass = (Class *)symtab->defs;
		 nClasses;
		 nClasses--, pClass++) {
		class = warpAddr(abfd, *pClass);
		NXMapRemove(classMap, warpAddr(abfd, class->name));
		removeMethods(abfd, warpAddr(abfd, class->methods));
		class = warpAddr(abfd, class->isa);
		removeMethods(abfd, warpAddr(abfd, class->methods));
	    }
	    for (nCats = symtab->cat_def_cnt, pCat = (Category *)pClass;
		 nCats;
		 nCats--, pCat++) {
		cat = warpAddr(abfd, *pCat);
		removeMethods(abfd, warpAddr(abfd, cat->instance_methods));
		removeMethods(abfd, warpAddr(abfd, cat->class_methods));
	    }
	}
    }
}

void printObjectCommand(char *args)
{
    static value streamVal = NULL, zeroVal = NULL, printSelVal = NULL;
    static int printSel = 0;
    static NXStream *hisStream = NULL;
    value objectVal;
    id hisObject;
    int len;
    
    if (!hisStream) {
	value openFunc, args[3];
	zeroVal = value_from_long(builtin_type_long, 0);
	release_value(zeroVal);
        openFunc = value_of_function("NXOpenMemory");
	args[0] = args[1] = zeroVal;
	args[2] = value_from_long(builtin_type_long, NX_WRITEONLY);
	streamVal = target_call_function(openFunc, 3, args);
	hisStream = (NXStream *)value_as_long(streamVal);
	printSel = lookup_child_selector("printForDebugger:");
	printSelVal = value_from_long(builtin_type_long, printSel);
	release_value(printSelVal);
	release_value(streamVal);
    } else {
	value seekFunc = value_of_function("NXSeek"),
	      args[] = {streamVal, zeroVal, zeroVal};
             
	target_call_function(seekFunc, 3, args);
    }
    objectVal = evaluate_expression(parse_c_expression(args));
    hisObject = (id)value_as_long(objectVal);
    if (hisObject) {
	unsigned imp = find_implementation(hisObject, (SEL)printSel);
	value printVal = value_from_long(builtin_type_long, imp),
	      args[] = {objectVal, printSelVal, streamVal};
	NXStream *myStream;
	char *myBuf, *myString;

	target_call_function(printVal, 3, args);
	
	myStream = WARP(hisStream);
	len = myStream->buf_ptr - myStream->buf_base;
	myBuf = [regionManager pointerFor: myStream->buf_base
	                         withSize: len];
	fwrite(myBuf, 1, len, stdout);
	fputs("\n", stdout);
    }
}

#define    TRIM    0177            /* Mask to strip quote bit */

static int globMatch(char *s, char *p)
{
    int scc;
    int ok, lc;
    int c, cc;

    for (;;) {
	scc = *s++ & TRIM;
	switch (c = *p++) {
    
	    case '[':
		ok = 0;
		lc = 077777;
		while (cc = *p++) {
		    if (cc == ']') {
			if (ok)
			    break;
			return (0);
		    }
		    if (cc == '-') {
			if (lc <= scc && scc <= *p++)
			    ok++;
		    } else
			if (scc == (lc = cc))
			    ok++;
		}
		if (cc == 0)
		    return 0;
		continue;
    
	    case '*':
		if (!*p)
		    return (1);
		for (s--; *s; s++)
		    if (globMatch(s, p))
			return (1);
		return (0);
    
	    case 0:
		return (scc == 0);
    
	    case '?':
		if (scc == 0)
		    return (0);
		continue;
    
	    default:
		if ((c & TRIM) != scc)
		    return (0);
		continue;
    
	}
    }
}

static int
pstrcmp(const void *ps1, const void *ps2)
{ return strcmp(*(char **)ps1, *(char **)ps2); }

static void
printNames(char **found, int numFound)
{
    int i, maxLen = 0, len, numPerLine, j;

    for (i = 0; i < numFound; i++) {
	len = strlen(found[i]);
	if (len > maxLen)
	    maxLen = len;
    }
    numPerLine = 80 / (maxLen + 1);
    if (numPerLine < 1) numPerLine = 1;
    if (numPerLine > numFound) numPerLine = numFound;
    maxLen += ((80 - (numPerLine * (maxLen + 1))) / numPerLine);
    qsort(found, numFound, sizeof(*found), pstrcmp);
    
    for (i = 0; i < numFound;) {
	for (j = 0; (i < numFound) && (j < numPerLine); j++, i++) {
	    printf("%-*s ", maxLen, found[i]);
	}
	printf("\n");
    }
}

static void
infoClassesCommand(char *args)
{
    int numAlloced = 24, numFound = 0, i, maxLen = 0, len, numPerLine, j,
        extra;
    char **found = malloc(numAlloced * sizeof(*found));
    NXMapState state;
    Class class;
    char *className;
    
    if (!args)
	args = "*";
    
    state = NXInitMapState(classMap);
    while (NXNextMapState(classMap, &state,
                          &(void *)className, &(void *)class)) {
	if (globMatch(className, args)) {
	    if (numFound == numAlloced) {
		numAlloced *= 2;
		found = realloc(found, numAlloced * sizeof(*found));
	    }
	    found[numFound++] = className;
	}
    }
    if (numFound) {
	printf("The following classes match %s.\n", args); 
	printNames(found, numFound);
    } else
	printf("No classes match %s.\n", args);
    free(found);
}

static void
infoSelectorsCommand(char *args)
{
    int numAlloced = 24, numFound = 0, i, maxLen = 0, len, numPerLine, j;
    char **found = malloc(numAlloced * sizeof(*found)), *selName;
    NXHashState state;
    
    if (!args)
	args = "*";
    
    state = NXInitHashState(selectorHash);
    while (NXNextHashState(selectorHash, &state, &(void *)selName)) {
	if (globMatch(selName, args)) {
	    if (numFound == numAlloced) {
		numAlloced *= 2;
		found = realloc(found, numAlloced * sizeof(*found));
	    }
	    found[numFound++] = selName;
	}
    }
    if (numFound) {
	printf("The following selectors match %s.\n", args); 
	printNames(found, numFound);
    } else
	printf("No selectors match %s.\n", args);
    free(found);
}
    
void
removeObjC(bfd *abfd)
{
    asection *modSect;
    if (modSect = bfd_get_section_by_name(abfd, "__OBJC__module_info"))
	removeModSect(abfd, modSect);
}

void _initialize_objc()
{
    add_com("print-object", class_vars, printObjectCommand,
            "Print object by sending \"printForDebugger:\" to it.\n");
    add_com_alias("po", "print-object", class_vars, 2);
    add_info("classes", infoClassesCommand, "All Objective C classes");
    add_info("selectors", infoSelectorsCommand, "All Objective C selectors");
}
