/*
 * Create and initialized a user task from an in-memory Mach-O immage.
 */
#import <varargs.h>
#import <mach/mach.h>
#import <mach-o/loader.h>

#define mach_port_t port_t
#define mach_task_self task_self

static kern_return_t launch_boot_program(
	struct mach_header	*mhp,
	unsigned int		arg_count,
	unsigned int		arg_len,
	va_list			args);
kern_return_t load_boot_program(
	mach_port_t		my_task,
	struct mach_header	*mhp);
static kern_return_t load_segment(
	mach_port_t		user_task,
	struct segment_command	*scp,
	vm_offset_t		file_base);
static kern_return_t load_thread(
	mach_port_t		user_task,
	struct thread_command	*tcp);

int launchverbose = 0;
 
mach_port_t		user_task;

kern_return_t boot_load_program(int va_alist, struct mach_header *mach_hdr) 
{
	register kern_return_t	result = KERN_SUCCESS;
	va_list			argv_ptr;
	char *			arg_ptr;
	unsigned int		arg_len = 0;
	unsigned int		arg_count;
	unsigned int timeoutflg;

	/*
	 * Build argument string.
	 * Ignore name argument.
	 */
	va_start(argv_ptr);
	for (;;) {
		arg_ptr = va_arg(argv_ptr, char *);
		if (!arg_ptr)
			break;
		arg_count++;
		arg_len += strlen(arg_ptr) + 1;	/* space for '\0' */
	}
	va_end(argv_ptr);

	/*
	 * Add space for:
	 *    arg_count
	 *    pointers to arguments
	 *    trailing 0 pointer
	 *    dummy 0 pointer to environment variables
	 *    and align to integer boundary
	 */
	arg_len += sizeof(int) + (2 + arg_count) * sizeof(char *);
	arg_len = (arg_len + (sizeof(int) - 1)) & ~(sizeof(int)-1);

	/*
	 * Ignore name argument.
	 */
	/*
	 * Launch user program.
	 */
//	en_release();
	va_start(argv_ptr);
	result = launch_boot_program(mach_hdr, arg_count, arg_len, argv_ptr);
	va_end(argv_ptr);
	if (result != KERN_SUCCESS) {
		printf("launch failed\n");
		return result;
	}


	printf("\n tftp_init restarting \n");
//	en_reattach();



	return result;
}


static kern_return_t launch_boot_program(
	struct mach_header	*mhp,
	unsigned int		arg_count,
	unsigned int		arg_len,
	va_list			args)
{
	register kern_return_t	result;
	mach_port_t		user_thread;
	mach_port_t		bootstrap_port;
	mach_port_t		my_task = mach_task_self();

	/*
	 * Create the user task to run the boot program.
	 */
	if (launchverbose)
		printf("Create task\n");
	result = task_create(my_task, FALSE, &user_task);
	if (result != KERN_SUCCESS)
		return result;

	/*
	 * Load Mach-O from physical memory into the task.  Create threads
	 * as listed in the Mach-O.
	 */
	result = load_boot_program(user_task, mhp);
	if (result != KERN_SUCCESS)
		return result;

	/*
	 * FIXME: figure out how to provide initial arguments to the task.
	 */

	return KERN_SUCCESS;
}

kern_return_t load_boot_program(
	mach_port_t		user_task,
	struct mach_header	*mhp)
{
	register kern_return_t	result;
	struct load_command *lcp;
	unsigned int ncmds = mhp->ncmds;
	unsigned int nthreads = 0;

	if (mhp->magic != MH_MAGIC || mhp->filetype != MH_EXECUTE)
		return KERN_FAILURE;

	/*
	 * Handle all segments of the Mach-O.
	 */
	for (  lcp = (struct load_command *)(mhp+1), result = KERN_SUCCESS
	     ; ncmds && result == KERN_SUCCESS;
	       ncmds--, lcp = (struct load_command *)((int)lcp + lcp->cmdsize))
	{
		if (launchverbose)
			printf("seg cmd 0x%x\n", lcp->cmd);
		switch (lcp->cmd) {
		case LC_SEGMENT:
			result = load_segment(user_task,
				(struct segment_command *)lcp,
				(vm_offset_t)mhp);
			break;

		case LC_THREAD:
			result = load_thread(user_task,
				(struct thread_command *)lcp);
			break;

		case LC_SYMTAB:	// Ignore the symbol table
			break;

		default:	// Bad file format.
			result = KERN_FAILURE;
			break;
		}
	}

	return result;
}

static kern_return_t load_segment(
	mach_port_t		user_task,
	struct segment_command	*scp,
	vm_offset_t		file_base)
{
	register kern_return_t	result;
	vm_offset_t		addr;
	vm_size_t		size;

	/*
	 * Ensure that the data's properly page-aligned.
	 */
	if (   scp->vmaddr != round_page(scp->vmaddr)
	    || scp->vmsize != round_page(scp->vmsize)
	    || scp->fileoff != round_page(scp->fileoff))
//	    || scp->filesize != round_page(scp->filesize))
		return KERN_MEMORY_ERROR;

	printf("\tsegment %s %d (%d+%d) bytes at 0x%x\n",
		scp->segname, scp->vmsize, scp->filesize,
		scp->vmsize-scp->filesize, scp->vmaddr);

	/*
	 * Allocate space in the user's task to write into.
	 */
	result = vm_allocate(user_task, &scp->vmaddr, scp->vmsize, FALSE);
	if (result != KERN_SUCCESS)
		return result;

	if (launchverbose)
		printf("Allocated space in target address space\n");
	if (scp->filesize) {
		addr = file_base + scp->fileoff;
		/*
		 * Write this data into the user's task.
		 */
		if (scp->filesize) {
			result = vm_write(user_task, scp->vmaddr, addr,
				scp->filesize);
			if (result != KERN_SUCCESS)
				return result;
		}
		if (launchverbose)
			printf("Wrote into space in target address space\n");

	}

	if (launchverbose)
		printf("Set protection max 0x%x init 0x%x \n", scp->maxprot,
		scp->initprot);

	/*
	 * Set protection on the user's memory range.
	 */
	(void) vm_protect(user_task,
			      scp->vmaddr,
			      scp->vmsize,
			      TRUE,
			      scp->maxprot);
	(void) vm_protect(user_task,
			      scp->vmaddr,
			      scp->vmsize,
			      FALSE,
			      scp->initprot);
	if (launchverbose)
		printf("Return from load segment\n");

	return KERN_SUCCESS;
}

static kern_return_t load_thread(
	mach_port_t		user_task,
	struct thread_command	*tcp)
{
	register kern_return_t	result;
	mach_port_t		user_thread;
	unsigned long		size;
	unsigned long		total_size;
	unsigned long		*ts = (unsigned long *)(tcp+1);
	int			flavor;

	result = thread_create(user_task, &user_thread);
	if (result != KERN_SUCCESS)
		return result;

	total_size = tcp->cmdsize - sizeof(struct thread_command);
	while (total_size > 0) {
		flavor = *ts++;
		size = *ts++;
		total_size -= (size+2)*sizeof(unsigned long);
		if (launchverbose){
			printf("Thread flavor %d size 0x%x total_size 0x%x\n",
			       flavor, size, total_size);
			printf(" dreg 0x%x 0x%X 0x%X 0x%x \n",
			*ts, *(ts +1), *(ts +2), *(ts +3));
			printf(" dreg 0x%x 0x%X 0x%X 0x%x \n",
			*(ts + 4), *(ts +5), *(ts +6), *(ts +7));
			printf(" areg 0x%x 0x%X 0x%X 0x%x \n",
			*(ts + 8), *(ts +9), *(ts +10), *(ts +11));
			printf(" areg 0x%x 0x%X 0x%X 0x%x \n",
			*(ts + 12), *(ts +13), *(ts +14), *(ts +15));
			printf(" sr 0x%x pc 0x%x  \n", *(ts + 16), *(ts + 17));
		}
		if (total_size < 0)
			return KERN_FAILURE;
		result = thread_set_state(user_thread, flavor,
			(thread_state_t)ts, size);
		if (result != KERN_SUCCESS)
			return result;
		ts += size;
	} 

	if (launchverbose)
		printf("Starting thread\n");

	return thread_resume(user_thread);
}
